/*
 * Copyright (c) 2016, The Linux Foundation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "libtrace/kbuffer.h"
#include "ras-arm-handler.h"
#include "ras-record.h"
#include "ras-logger.h"
#include "ras-report.h"
#include "ras-cpu-isolation.h"

#ifdef HAVE_CPU_FAULT_ISOLATION
static void trace_print_hex(struct trace_seq *s, const uint8_t *buf, int buf_len)
{
	for (int i = 0; i < buf_len; ++i) {
		trace_seq_printf(s, "%2.2x", buf[i]);
	}
}

static int is_core_failure(unsigned long value)
{
	/* 
	 * core failure:
	 * Bit 0\1\3: (at lease 1)
	 * Bit 2: 0
	 */
	return (value & 0xf) && !(value & (0x1 << 2));
}

static int count_errors(struct event_format *event, const uint8_t *data, int len)
{
	/*
	 * According to UEFI_2_9_2021_03_18 specification chapter N2.4.4,
	 * the length of struct processor error information is 32, the byte
	 * length of the Flags field is 1, and the byte offset is 7 in the struct.
	 */
	int cur_offset = 7;
	unsigned long value;
	int num = 0;
	if (len % PEI_ERR_SIZE != 0) {
		log(TERM, LOG_ERR, "the event data does not match to the ARM Processor Error Information Structure\n");
		return num;
	}
	while (cur_offset < len) {
		value = pevent_read_number(event->pevent, data+cur_offset, FLAGS_SIZE);
		if (is_core_failure(value)) {
			num++;
			log(TERM, LOG_INFO, "Error in cpu core catched\n");
		}
		cur_offset += PEI_ERR_SIZE;
	}
	return num;
}
#endif

int ras_arm_event_handler(struct trace_seq *s,
			 struct pevent_record *record,
			 struct event_format *event, void *context)
{
	unsigned long long val;
	struct ras_events *ras = context;
	time_t now;
	struct tm *tm;
	struct ras_arm_event ev;

	memset(&ev, 0, sizeof(ev));

	/*
	 * Newer kernels (3.10-rc1 or upper) provide an uptime clock.
	 * On previous kernels, the way to properly generate an event would
	 * be to inject a fake one, measure its timestamp and diff it against
	 * gettimeofday. We won't do it here. Instead, let's use uptime,
	 * falling-back to the event report's time, if "uptime" clock is
	 * not available (legacy kernels).
	 */

	if (ras->use_uptime)
		now = record->ts/user_hz + ras->uptime_diff;
	else
		now = time(NULL);

	tm = localtime(&now);
	if (tm)
		strftime(ev.timestamp, sizeof(ev.timestamp),
			 "%Y-%m-%d %H:%M:%S %z", tm);
	trace_seq_printf(s, "%s\n", ev.timestamp);

	if (pevent_get_field_val(s, event, "affinity", record, &val, 1) < 0)
		return -1;
	ev.affinity = val;
	trace_seq_printf(s, " affinity: %d", ev.affinity);

	if (pevent_get_field_val(s, event, "mpidr", record, &val, 1) < 0)
		return -1;
	ev.mpidr = val;
	trace_seq_printf(s, "\n MPIDR: 0x%llx", (unsigned long long)ev.mpidr);

	if (pevent_get_field_val(s, event, "midr", record, &val, 1) < 0)
		return -1;
	ev.midr = val;
	trace_seq_printf(s, "\n MIDR: 0x%llx", (unsigned long long)ev.midr);

	if (pevent_get_field_val(s, event, "running_state", record, &val, 1) < 0)
		return -1;
	ev.running_state = val;
	trace_seq_printf(s, "\n running_state: %d", ev.running_state);

	if (pevent_get_field_val(s, event, "psci_state", record, &val, 1) < 0)
		return -1;
	ev.psci_state = val;
	trace_seq_printf(s, "\n psci_state: %d", ev.psci_state);

#ifdef HAVE_CPU_FAULT_ISOLATION
	int cpu;
	if (pevent_get_field_val(s, event, "cpu", record, &val, 1) < 0)
		return -1;
	cpu = val;
	trace_seq_printf(s, "\n cpu: %d", cpu);

	/* record cpu error */
	if (pevent_get_field_val(s, event, "sev", record, &val, 1) < 0)
		return -1;
	/* refer to UEFI_2_9_2021_03_18 specification chapter N2.2 Table N-5 */
	switch (val) {
	case GHES_SEV_NO:
		ev.severity = "Informational";
		break;
	case GHES_SEV_CORRECTED:
		ev.severity = "Corrected";
		break;
	case GHES_SEV_RECOVERABLE:
		ev.severity = "Recoverable";
		break;
	default:
	case GHES_SEV_PANIC:
		ev.severity = "Fatal";
	}
	trace_seq_printf(s, "\n severity: %s", ev.severity);

	if (val == GHES_SEV_CORRECTED || val == GHES_SEV_RECOVERABLE) {
		int len, nums;
		ev.error_info = pevent_get_field_raw(s, event, "buf", record, &len, 1);
		if (!ev.error_info)
			return -1;
		ev.length = len;
		trace_seq_printf(s, "\n processor_err_info: ");
		trace_print_hex(s, ev.error_info, len);
		/* relate to enum error_type */
		nums = count_errors(event, ev.error_info, len);
		if (nums > 0) {
			struct error_info err_info = {nums, now, val};
			ras_record_cpu_error(&err_info, cpu);
		}
	}
#endif

	/* Insert data into the SGBD */
#ifdef HAVE_SQLITE3
	ras_store_arm_record(ras, &ev);
#endif

#ifdef HAVE_ABRT_REPORT
	/* Report event to ABRT */
	ras_report_arm_event(ras, &ev);
#endif

	return 0;
}
